//
// fraktal - visualization of Mandelbrot and Julia sets
// Copyright (C) 1999-2020  Joachim Reichel <joachim.reichel@posteo.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#include <cassert>
#include <cmath>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <iostream>

#include "compute.h"

#if !defined (HAVE_CUDA)
#if defined (HAVE_OPENMP)
#  include <omp.h>
#endif // HAVE_OPENMP
#endif // HAVE_CUDA

const int g_bailout    = 2;
const int g_block_size = 16;

#if defined (HAVE_CUDA)
__host__ __device__
#endif // HAVE_CUDA
double mandelbrot_set( double a, double b, int kmax)
{
    double u = 4 * (a*a + b*b);
    double v = u - 2*a + 0.25;

    if( u + 8*a + 3.75 < 0)
        return 0;
    if( v - sqrt(v) + 2*a - 0.5 < 0)
        return 0;

    int k = 0;
    double x1 = a;
    double y1 = b;
    double d;

    do {
        double x2 = x1*x1 - y1*y1 + a;
        double y2 = 2 * x1 * y1 + b;
        d  = x2*x2 + y2*y2;
        x1 = x2;
        y1 = y2;
        k++;
    } while( (d <= g_bailout*g_bailout) && (k < kmax));

    if( k >= kmax)
        return 0;

    return k + log( log( static_cast<double>( g_bailout*g_bailout)) / log( d)) / log( 2.0);
}

#if defined (HAVE_CUDA)
__host__ __device__
#endif // HAVE_CUDA
double inverse_mandelbrot_set( double a, double b, int kmax)
{
    if( (a == 0.0) && (b == 0.0))
        a = 1E-10;

    double denom = a*a + b*b;
    a = 0.25 * a / denom - 1;
    b = 0.25 * b / denom;

    return mandelbrot_set( a, b, kmax);
}

#if defined (HAVE_CUDA)
__host__ __device__
#endif // HAVE_CUDA
double julia_set( double x1, double y1, int kmax)
{
    int k = 0;
    double d;

    do {
        double x2 = x1*x1 - y1*y1 + 0.11;
        double y2 = 2 * x1 * y1 - 0.66;
        d  = x2*x2 + y2*y2;
        x1 = x2;
        y1 = y2;
        k++;
    } while( (d <= g_bailout*g_bailout) && (k < kmax));

    return k + log( log( static_cast<double>( g_bailout*g_bailout)) / (log( d)) ) / log( 2.0);
}

#if defined (HAVE_CUDA)
__host__ __device__
#endif // HAVE_CUDA
void value_to_rgb( double v, double& r, double& g, double& b)
{
    if( v == 0.0) {

        r = g = b = 0.0;

    } else {

        v = v / 200.0;
        v = v - floor( v);

        if( v < 0.25) {
            double x = 4.0*v;
            r = x;
            g = x;
            b = 0.52 + 0.44*x;
        } else if( v < 0.5) {
            double x = 4.0*v - 1.0;
            r = 1.0;
            g = 1.0  - 0.29*x;
            b = 0.96 - 0.96*x;
        } else if( v < 0.75) {
            double x = 4.0*v - 2.0;
            r = 1.0  - 0.39*x;
            g = 0.71 - 0.71*x;
            b = 0.0;
        } else {
            double x = 4.0*v - 3.0;
            r = 0.61 - 0.61*x;
            g = 0.0;
            b = 0.52*x;
        }
    }
}

#if defined (HAVE_CUDA)
__global__
void kernel( float* result, int w, int h, double cx, double cy, double s, Function f, int quality, int stride)
{
    int samples = quality >= 0 ? 1 << quality : 1;
    int samples_squared = samples * samples;
    int size = quality <= 0 ? 1 << -quality : 1;
    int kmax = static_cast<int>( 400 * (1 + log( s) / log( 10.0)));

    const int x = size * (blockIdx.x * blockDim.x + threadIdx.x);
    const int y = size * (blockIdx.y * blockDim.y + threadIdx.y);
    if( x >= w || y >= h)
        return;

    // compute extent in world coordinates
    double dx = (w < h) ? 1/s : 1/s * w/h;
    double dy = (h < w) ? 1/s : 1/s * h/w;
    // compute lower left corner in world coordinates
    double px = cx - dx;
    double py = cy - dy;

    double red   = 0;
    double green = 0;
    double blue  = 0;

   for( int i = 0; i < samples; ++i)
        for( int j = 0; j < samples; ++j) {

            // compute function value
            double a = px + (x + (i + 0.5)/samples)/w * 2*dx;
            double b = py + (y + (j + 0.5)/samples)/h * 2*dy;
            double value;
            switch( f) {
                case FCT_MANDELBROT:         value = mandelbrot_set        ( a, b, kmax); break;
                case FCT_INVERSE_MANDELBROT: value = inverse_mandelbrot_set( a, b, kmax); break;
                case FCT_JULIA:              value = julia_set             ( a, b, kmax); break;
                default:                     value = 0.0;
            }

            // convert function value to RGB color
            double r, g;
            value_to_rgb( value, r, g, b);

            // add RGB color to previous samples
            red   += r;
            green += g;
            blue  += b;
        }

    // compute average RGB color of all samples
    red   /= samples_squared;
    green /= samples_squared;
    blue  /= samples_squared;

    // store RGB color in result
    for( int i = 0; (i < size) && (x+i < w); i++)
        for( int j = 0; (j < size) && (y+j < h); j++) {
             int offset       = 4 * ((y+j)*stride + x+i);
             result[offset  ] = red;
             result[offset+1] = green;
             result[offset+2] = blue;
             result[offset+3] = 1.0;
        }
}

#define check(code) __check (code, __FILE__, __LINE__)

void __check( cudaError code, const char* file, int line)
{
    if( cudaSuccess != code) {
        std::cout << file << ":" << line << ": " << cudaGetErrorString( code) << std::endl;
        exit( 1);
    }
}

void compute_cuda( float* h_result, int w, int h, double cx, double cy, double s, Function f, int quality, int stride)
{
    assert( h_result);
    assert( w > 0);
    assert( h > 0);
    assert( s > 0);
    assert( stride >= w);

    float* d_result;
    check( cudaMalloc( &d_result, 4*stride*h*sizeof( float)));

    int size = quality <= 0 ? 1 << -quality : 1;
    int w2 = (w+size-1)/size;
    int h2 = (h+size-1)/size;

    dim3 grid( (w2+g_block_size-1)/g_block_size, (h2+g_block_size-1)/g_block_size, 1);
    dim3 threads( g_block_size, g_block_size, 1);
    kernel<<< grid, threads >>>( d_result, w, h, cx, cy, s, f, quality, stride);

    check( cudaMemcpy( h_result, d_result, 4*stride*h*sizeof( float), cudaMemcpyDeviceToHost));
    check( cudaFree( d_result));
}
#else
void compute_cuda( float* /*h_result*/, int /*w*/, int /*h*/, double /*cx*/, double /*cy*/, double /*s*/, Function /*f*/, int /*quality*/, int /*stride*/)
{
    assert( false);
}
#endif // HAVE_CUDA

void compute_cpu( float* result, int w, int h, double cx, double cy, double s, Function f, int quality, int stride)
{
    assert( w > 0);
    assert( h > 0);
    assert( s > 0);
    assert( stride >= w);

    // compute extend in world coordinates
    double dx = (w < h) ? 1/s : 1/s * w/h;
    double dy = (h < w) ? 1/s : 1/s * h/w;
    // compute lower left corner in world coordinates
    double px = cx - dx;
    double py = cy - dy;

    int samples = quality >= 0 ? 1 << quality : 1;
    int samples_squared = samples * samples;
    int size = quality <= 0 ? 1 << -quality : 1;
    int kmax = static_cast<int>( 400 * (1 + log( s) / log( 10.0)));

#if defined (HAVE_OPENMP)
#  pragma omp parallel for
#endif // HAVE_OPENMP
    for( int y = 0; y < h; y += size) {
        for( int x = 0; x < w; x += size) {

            double red   = 0;
            double green = 0;
            double blue  = 0;

            for( int i = 0; i < samples; ++i)
                for( int j = 0; j < samples; ++j) {

                    // compute function value
                    double a = px + (x + (i + 0.5)/samples)/w * 2*dx;
                    double b = py + (y + (j + 0.5)/samples)/h * 2*dy;
                    double value;
                    switch( f) {
                        case FCT_MANDELBROT:         value = mandelbrot_set        ( a, b, kmax); break;
                        case FCT_INVERSE_MANDELBROT: value = inverse_mandelbrot_set( a, b, kmax); break;
                        case FCT_JULIA:              value = julia_set             ( a, b, kmax); break;
                        default:                     value = 0.0;
                    }

                    // convert function value to RGB color
                    double r, g;
                    value_to_rgb( value, r, g, b);

                    // add RGB color to previous samples
                    red   += r;
                    green += g;
                    blue  += b;
                }

            // compute average RGB color of all samples
            red   /= samples_squared;
            green /= samples_squared;
            blue  /= samples_squared;

            // store RGB color in result
            for( int i = 0; (i < size) && (x+i < w); i++)
                for( int j = 0; (j < size) && (y+j < h); j++) {
                    int offset       = 4 * ((y+j)*stride + x+i);
                    result[offset  ] = red;
                    result[offset+1] = green;
                    result[offset+2] = blue;
                    result[offset+3] = 1.0;
                }
        }
    }
}
