//
// fraktal - visualization of Mandelbrot and Julia sets
// Copyright (C) 1999-2020  Joachim Reichel <joachim.reichel@posteo.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#ifndef FRAKTAL_RENDERER_H
#define FRAKTAL_RENDERER_H

#include <string>

#if defined( HAVE_GLEW)
#  include <GL/glew.h>
#endif // HAVE_GLEW

#include <GL/gl.h>

#include "compute.h"

/// Mode to be used for visualization.
enum Mode {
  MODE_CPU,     ///< Use the CPU
  MODE_OPENGL,  ///< Use OpenGL
  MODE_CUDA     ///< Use CUDA
};

/// This class provides a renderer for fractal images.
///
/// The renderer visualizes several fractal image functions.
///
/// The section of the image is defined by a center (cx, cy) and scaling factor
/// s in world coordinates. For a window of width w and height h the image
/// section is given by
///
///   [ cx - w/h/s, cx + w/h/s ] x [ cy - 1/s, cy + 1/s ] if w >= h, and
///   [ cx - 1/s, cx + 1/s ] x [ cy - h/w/s, cy + h/w/s ] if w <= h.
///
/// The section can be altered by translation and scaling operations.
class Renderer
{
public:

    /// Constructor.
    Renderer();

    /// Destructor.
    ~Renderer();

    /// Initializes the OpenGL context.
    void init();

    /// Draws the image in the current OpenGL context.
    void draw( int width, int height);

    /// Resizes the viewport to width \p w and height \a h.
    void resize( int w, int h) const;

#if defined( HAVE_PNG) || defined( HAVE_JPEG)
    /// Saves the image of width \p w and height \a h to filea filename
    void screenshot( int width, int height, const std::string& filename);
#endif // HAVE_PNG || HAVE_JPEG

    /// Resets the transformation.
    void reset();

    /// Translates center by (\p cx, \a cy).
    void translate( double dx, double dy);

    /// Scales by \p s around current center.
    void scale( double s);

    /// Scales by \p s around (\a px, \a py).
    void scale( double s, double px, double py);


    /// Sets the quality level to \p q.
    ///
    /// On quality level \p q <= 0, the image is composed of pixels of size
    /// 2^(-q) times 2^(-q) with 1 sample each. On quality level \p q >= 0, the
    /// image is composed of pixels of size 1x1 with 2^q x 2^q samples.
    void set_quality( int q);

    /// Returns the quality level.
    int get_quality() const;

    /// Sets the image center to (\p cx, \a cy).
    void set_center( double cx, double cy);

    /// Returns image center as (\p cx, \a cy).
    void get_center( double& cx, double& cy) const;

    /// Sets the scaling factor to \p s.
    void set_scaling( double s);

    /// Returns the scaling factor as \p s.
    double get_scaling() const;

    /// Sets the function to be visualized to \p f.
    void set_function( Function f);

    /// Returns the function to be visualized.
    Function get_function() const;

    /// Sets the mode to be used to \p m.
    void set_mode( Mode m);

    /// Returns the mode to be used.
    Mode get_mode() const;


    /// Returns information about current center, scaling factor, etc.
    std::string status( int width, int height) const;

private:

#if defined( HAVE_JPEG)
    /// Saves the image as JPEG file.
    void write_jpeg_file( const char* buffer, int width, int height, const std::string& filename) const;
#endif // HAVE_JPEG

#if defined( HAVE_PNG)
    /// Saves the image as PNG file.
    void write_png_file( const char* buffer, int width, int height, const std::string& filename) const;
#endif // HAVE_PNG


#if defined( HAVE_GLEW)
    /// Enables OpenGL shaders
    void enable_shaders();

    /// Switches the OpenGL shader to a different function.
    void switch_shaders();

    /// Disables OpenGL shaders.
    void disable_shaders();

    /// Prints OpenGL shader debug messages
    void print_shader_info_log( GLuint obj) const;

    /// Prinst OpenGL program debug messages
    void print_program_info_log( GLuint obj) const;
#endif // HAVE_GLEW


    /// The x-coordinate of center.
    double m_cx;

    /// The y-coordinate of center.
    double m_cy;

    /// The scaling factor.
    double m_s;

    // The quality level.
    int m_quality;

    /// The function to be visualized.
    Function m_function;

    /// The mode to be used.
    Mode m_mode;

    /// Ondicates whether OpenGL shaders are actually enabled.
    bool m_shaders_enabled;


#if defined( HAVE_GLEW)
    /// Maximum number of iterations.
    GLfloat m_kmax;

    /// Currently active vertex shader.
    GLuint m_vertex_shader;

    /// Currently active fragment shader.
    GLuint m_fragment_shader;

    /// Currently active shader program.
    GLuint m_shaderprogram;

    /// Fragment shader for the mandelbrot function.
    GLuint m_mandelbrot_shader;

    /// Fragment shader for the inverse mandelbrot function.
    GLuint m_inverse_mandelbrot_shader;

    /// Fragment shader for the julia function.
    GLuint m_julia_shader;

    /// Fragment shader corresponding to \c m_function.
    GLuint m_shader;
#endif // HAVE_GLEW
};

#endif // FRAKTAL_RENDERER_H
